#include <string.h>
#include <list>
using namespace std;
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <netdb.h>
#include <ctype.h>
#include <time.h>

#include <OpenIPMI/ipmiif.h>
#include <OpenIPMI/ipmi_smi.h>
#include <OpenIPMI/ipmi_err.h>
#include <OpenIPMI/ipmi_auth.h>
#include <OpenIPMI/ipmi_lan.h>
#include <OpenIPMI/ipmi_posix.h>

#define MAX_SENSOR_NAME_SIZE 128
static os_handler_t *os_hnd;

typedef struct _rwwsensor {
  char shortname[MAXSENSORNAME];
  char name[MAXSENSORNAME];
  double critlow,alertlow,alerthigh,crithigh;
} rwwsensor;

typedef struct sdata_s
{
    unsigned int       refcount;
    ipmi_sensor_id_t   sensor_id;
    char               name[MAX_SENSOR_NAME_SIZE];
    ipmi_event_state_t *es;
    ipmi_thresholds_t  *th;
    int                state_sup;
    int                thresh_sup;

    struct sdata_s     *next, *prev;
} sdata_t;


  
class chassis{
 public:
  char name[MAXCHASSISNAME], *fanname[MAXFANS], *tempname[MAXTEMPS], *voltname[MAXVOLTS];
  char *fanshortname[MAXFANS], *tempshortname[MAXTEMPS], *voltshortname[MAXVOLTS];
  int numfans,numtemps,numvolts;
  double templimit[MAXTEMPS][4], voltlimit[MAXVOLTS][4], fanlimit[MAXFANS][4];

  chassis() {  //initialize all variables
    int count;
    numfans=0; numtemps=0; numvolts=0;
    for (count=0;count<MAXFANS;count++) {
      fanname[count]=NULL;
      fanshortname[count]=NULL;
      fanlimit[count][0]=-1;
      fanlimit[count][1]=0;
      fanlimit[count][2]=40000;
      fanlimit[count][3]=40001;
    }
    for (count=0;count<MAXVOLTS;count++) {
      voltname[count]=NULL;
      voltshortname[count]=NULL;
      voltlimit[count][0]=-1;
      voltlimit[count][1]=0;
      voltlimit[count][2]=40000;
      voltlimit[count][3]=40001;
    }
    for (count=0;count<MAXTEMPS;count++) {
      tempname[count]=NULL;
      tempshortname[count]=NULL;
      templimit[count][0]=-1;
      templimit[count][1]=0;
      templimit[count][2]=40000;
      templimit[count][3]=40001;
    }
  }
};

class node{
 public:
  chassis *thischassis;
  node();
  node(chassis *chassisptr);
  bool getdata();
  bool command(int command);
  char *name, *username, *password;
  double volt[MAXVOLTS], temp[MAXTEMPS], fan[MAXFANS];
  int checklimits();
  bool ispowered(){return power;}  
  bool isdead(){return dead;}  
  void shutdown();
  bool registersensor(ipmi_sensor_t *sensor);
  bool reportsensor(char *sensorname, double value);
  bool reportdomain(int err);

  time_t lastupdate, lastquery, lastcheck;
  ipmi_con_t  *ipmicon;

  bool zerosensors();

 private:
  bool addsensor(ipmi_sensor_t *sensor);
  bool inShutdown, inAlert, dead, wasdead, updated, power;
  int processline(char *sensorline);
  
  ipmi_sensor_t *mysensors[MAXSENSORS];
  sdata_t *mysdatas[MAXSENSORS];

  int sensorcount;
  
  time_t firsttime;
};

void error_msg ( char *error );

class cluster{
 public:
//  cluster(char *clusterfilename, char *chassisfilename);
  bool init(char *clusterfilename, char *chassisfilename, char *commandfile);
  void process();
  bool command(char *name, int command);
  bool poweroffnode(char *name);
  bool registersensor(char *name, ipmi_sensor_t *sensor);
  bool reportsensor(char *name, char *sensorname, double value);
  bool reportdomain(char *name, int err);
  bool zerosensors(char *name);
  
  cluster() { 
    memset(emailto,MAXEMAILADDRESS,'\0');
    memset(emailfrom,MAXEMAILADDRESS,'\0');
    memset(emailserver,MAXEMAILSERVER,'\0');
    processalerts=false;
    apachegid=0;
    return;
  }

 private:
  time_t lastoutputtime;
  node mynodes[MAXNODES];
  int numnodes,numchassis;
   
  node *findnode(char *name);
  chassis *findchassis(char *name);
  void getdata();
  void checklimits();
  void outputdata();
  void outputautoconf();
  void parsechassis(char *);
  void validatechassis();
  void parsecluster(char *);
  void setupemail();
  char emailto[MAXEMAILADDRESS], emailfrom[MAXEMAILADDRESS];
  char emailserver[MAXEMAILSERVER];
  bool processalerts;
  void setupcommandfifo(char *filename);
  void processcommandfifo();
  int commandfifo;
  gid_t apachegid;
};


class alert {
 public:
  int type, index;
  long occurances;
  double value, limits[4];
  time_t timestamp, reported;   //when added, reported=0 and timestamp=current time.
  // at reporting time, look for all alerts that have timestamp>reported+ALERTFREQUENCY.
  // the +ALERTFREQUENCY makes it not report a problem again until that many more seconds have passed.
  // ALERTFREQUENCY is hard set in defines.hh to 15 minutes worth of seconds.  There is now a
  // soft frequency to group messages if sending a new alert about a different sensor, they all get grouped together into 1 email then



  char nodename[MAXNODENAME];
  bool operator<(const alert& T);   //this function will < by (timestamp-reported) to facilitate the above
//  bool operator==(const alert& T);  //used to compare type, index, and nodename only - other parts are ignored
                                    //used to determine if this is a new alert or an update to an existing one
  bool matches(int comptype, char compnodename[], int compindex);
  bool matches(char compnodename[]);
  alert(int mytype, char mynodename[], int myindex, double myvalue, double mylimits[]);
  int delaytime(bool sendingalready);
};
  
class email{
 public:
  bool clear(){
    bodyindex=body;
    memset(body,'\0',MAXEMAILSIZE);  //fill it will null terminating 0's
    return true;
  };
  bool add(alert i);

  void init(char toaddress[], char fromaddress[], char emailserver[]){
    strncpy(to,toaddress,MAXEMAILADDRESS);
    strncpy(from,fromaddress,MAXEMAILADDRESS);
    strncpy(server, emailserver, MAXEMAILSERVER);
    strncpy(subject,"Automatic alert from your Microway Cluster",MAXEMAILSUBJECT);
    body=(char *)malloc(MAXEMAILSIZE);
  }
  email(){bodyindex=body;}
  void *send();/* {
    //this will get forked off
    fprintf(ERROROUTPUT,"This is where I'd send the following email to %s from %s:\n%s\n",to,from,body);
    return true;
  }*/
  short int threadstatus;  //  see enum for ALERT_ status flags
 private:
  char *body; // used to be a static array, now malloced instead- [MAXEMAILSIZE];
  char to[MAXEMAILADDRESS], from[MAXEMAILADDRESS], server[MAXEMAILSERVER];
  char subject[MAXEMAILSUBJECT];
  char *bodyindex;
};

class alerts{
 public:
  void add(int type, char nodename[], int index, double value, double limits[]);
  void prune(int type, char nodename[], int index);
  void remove(char nodename[]);

   //add 1st looks in myalerts[] for an existing entry for this alert.  if it exists,
   //it is just updated with the current value and timestamp
   //if it doesn't exist, a new one is added with these values
  void process(); //checks if alerts are waiting and sends them, etc as needed
  void setupemail(char *emailto, char *emailfrom, char *emailserver) {
    myemail.init(emailto,emailfrom,emailserver);
    return;
  }
 private:
//  int find(int type, char nodename[], int index); //returns index to myalerts with the match, or -1 if no match
  list<alert> myalerts;
  email myemail;  
//  expire();  //called at the beginning of generateEmail to elimate outdated alerts before sending.  alerts that occured more than 30 mins ago are expired from the list
//  bool generateEmail(); //this gets called by the process function
  
};

struct shutdown_node {
  char nodename[MAXNODENAME], username[MAXNODENAME], password[MAXNODENAME];
  pthread_t thread;
  bool *inShutdown;
};
