#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <list>
#include <syslog.h>
#include <pthread.h>
extern int nodaemon;

int comparelimits(double value, double limits[]) {
  if (value < limits[CRITLOW]) return -2;
  else if (value < limits[ALERTLOW]) return -1;
  else if (value > limits[CRITHIGH]) return 2;
  else if (value > limits[ALERTHIGH]) return 1;
  else return 0;
}

int parsesensor(rwwsensor *thissensor, char *origline) {    //FIXME -this needs WAY more error checking
  char tempstring[120];
  bool inquote=false;
  int index, param;
  char *line=origline;
  memset(tempstring,120,'\0');
  for(param=0; param<6; param++){
    index=0;
    while( (inquote==true || *line!=' ') && *line!='\0' && *line!=EOF){
      if(*line=='"') {
        inquote==true ? inquote=false : inquote=true;
        line++;
      } else{
        tempstring[index]=*line;
        line++;
        index++;
      }
    }
    if(tempstring[0]==' ') { //ignore an extraneous space between params
      param--;
      continue;
    }
    if(tempstring[index-1]=='\n') tempstring[index-1]='\0';  //replace newline with EOS
    else tempstring[index]='\0'; //add the null terminator to the extracted string
    
    switch(param){
      case 0:
        strncpy(thissensor->shortname,tempstring,MAXSENSORNAME);
        break;
      case 1:
        strncpy(thissensor->name,tempstring,MAXSENSORNAME);
        break;
      case 2:
        thissensor->critlow=atof(tempstring);
        break;
      case 3:
        thissensor->alertlow=atof(tempstring);
        break;
      case 4:
        thissensor->alerthigh=atof(tempstring);
        break;
      case 5:
        thissensor->crithigh=atof(tempstring);
        break;
    }
    line++;
  }
  return true;
}

int parseemail(char *setting, char *origline) {
  char tempstring[120];
  bool inquote=false;
  int index, param;
  char *line=origline;
  memset(tempstring,120,'\0');
  
  for(param=0; param<1; param++){
    index=0;
    while( (inquote==true || *line!=' ') && *line!='\0' ){
      if(*line=='"') {
        inquote==true ? inquote=false : inquote=true;
        line++;
      } else{
        tempstring[index]=*line;
        line++;
        index++;
      }
    }
    if(tempstring[0]==' ') { //ignore an extraneous space between params
      param--;
      continue;
    }
    if(tempstring[index-1]=='\n') tempstring[index-1]='\0';  //replace newline with EOS
    else tempstring[index]='\0'; //add the null terminator to the extracted string

    strncpy(setting,tempstring,MAXEMAILADDRESS);
    line++;
  }
  return true;
}


int parsenode(char *nodename, char *chassisname, char *username, char *password, char *origline) {
//FIXME -this function needs WAY more error checking
  char tempstring[120];
  bool inquote=false;
  int index, param;
  char *line=origline;
  memset(tempstring,120,'\0');
  
  for(param=0; param<4; param++){
    index=0;
    while( (inquote==true || *line!=' ') && *line!='\0' ){
      if(*line=='"') {
        inquote==true ? inquote=false : inquote=true;
        line++;
      } else{
        tempstring[index]=*line;
        line++;
        index++;
      }
    }
    if(tempstring[0]==' ') { //ignore an extraneous space between params
      param--;
      continue;
    }
    if(tempstring[index-1]=='\n') tempstring[index-1]='\0';  //replace newline with EOS
    else tempstring[index]='\0'; //add the null terminator to the extracted string

    switch(param){
      case 0:
        strncpy(nodename,tempstring,MAXNODENAME);
        break;
      case 1:
        strncpy(chassisname,tempstring,MAXCHASSISNAME);
        break;
      case 2:
        strncpy(username,tempstring,MAXNODENAME);
        break;
      case 3:
        strncpy(password,tempstring,MAXCHASSISNAME);
        break;
    }
    line++;
  }
  return true;
}

void error_msg ( char *error ) {
  if ( nodaemon ) {
    fprintf ( stderr, "%s", error );
    fflush (stderr);
  }
  //else    // syslog should occur regardless of debug activity.
  syslog ( LOG_DAEMON | LOG_ERR, "%s\n", error );
}

void info_msg ( char *error ) {
  if ( nodaemon )
    fprintf ( stderr, "%s", error );
  // else  // syslog should occur regardless of debug activity.
  syslog ( LOG_DAEMON | LOG_INFO, "%s\n", error );
}


void *safeShutdown ( void *thenode) {
  char temp[500],temp2[500];
  shutdown_node *thisnode=(shutdown_node *)thenode;
  FILE *process;
  // canceltype set appropriates
  pthread_setcancelstate ( PTHREAD_CANCEL_ENABLE, NULL );
  pthread_setcanceltype ( PTHREAD_CANCEL_ASYNCHRONOUS, NULL );
  snprintf ( temp, 499, RSH" %s /sbin/halt &>/dev/null", thisnode->nodename );
  snprintf ( temp2, 499, "Safely shutting down %s with this command: %s\n", thisnode->nodename, temp);
  error_msg(temp2);
  

  process=popen(temp,"r");
  if ( process==NULL ) {
    error_msg((char *)"Failed to send rsh poweroff\n");
    pthread_exit(NULL);
    return NULL;
  } else {
    fgets(temp,500,process);
    //sprintf(temp2,"rsh popen output this: \"%s\"\n",temp);
    //error_msg(temp2);
    pclose(process);
    pendingalerts.remove(thisnode->nodename);   //remove pending alerts for this node - it's being turned off
    pthread_exit(NULL);
    return NULL;
  }
}

void *ensureShutdown (void *thenode) {
  char temp[500];
  shutdown_node *thisnode=(shutdown_node *)thenode;
  struct timespec mytimer;
  time_t curtime;

  sleep(10);  //sleep 10 seconds before even trying anything
  curtime=time(NULL);
  mytimer.tv_sec=curtime+60;  //wait up to one minute

  if(pthread_timedjoin_np( thisnode->thread, NULL, &mytimer)!=0) { //if the thread never exitted
    pthread_cancel(thisnode->thread); //kill the rsh thread
  }
  if( (time(NULL)-curtime) > 45 ) { //it took at least 45 seconds for the thread to join
    sleep(45); //sleep 45 more seconds, at least 90 total this way, but not more than 105
  } else {
    sleep (90 - (time(NULL)-curtime));    //sleep however many seconds more you need to
  }                                       //total 90 since the thread started
 
  sprintf(temp, "Completing auto-shutdown of %s with an IPMI power off to ensure it is completely down.\n",thisnode->nodename);
  error_msg(temp);
  mycluster.command(thisnode->nodename, POWEROFF); 
  *thisnode->inShutdown=false;
  pendingalerts.remove(thisnode->nodename);   //remove pending alerts for this node - it's being turned off
  free(thisnode); 
  pthread_exit(NULL);
  return NULL;

}
