#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <iostream>
#include <string.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <grp.h>
using namespace std;

#include "defines.hh"
#include "datastructs.hh"
#include "openipmi.hh"
#include "globals.hh"
#include "helpers.cc"
#include "nodes.cc"
#include "cluster.cc"
#include "alerts.cc"
#include "email.cc"
#include "openipmi.cc"

void print_usage () {
  fprintf (stderr, "\nUsage: ipmimon <Options>\n");

  fprintf (stderr,
            "  -h  --help\t\tDisplay this usage information\n"
            "  -d  --debug\t\tRun process in foreground, displaying debug info.\n"
            "  -p  --path <path>\tSpecify the absolute path to the data and command files.\n" );
  exit (2);
}


//////////// MAIN FUNCTION //
int main(int argc, char *argv[]) {

  int next_option, count;
  char const short_options[] = "dhp:";     // p: means p takes an argument.
  struct option long_options[] = {
    {"debug", 0, NULL, 'd'},
    {"help", 0, NULL, 'h'},
    {"path", 1, NULL, 'p'},
    {NULL, 0, NULL, 0}
  };

  strcpy(datadir, DEFAULTDATADIR);

  do {
    next_option =
      getopt_long ( argc, argv, short_options, long_options, NULL );
    switch ( next_option ) {
       case 'd':
         nodaemon++;
         break;
       case 'h':               //display usage and exit cleanly
         print_usage ();
         break;
       case 'p':               //change directory
         if ( (optarg == NULL) || (optarg[0] != '/') )  // must be absolute path
           print_usage ();
         strcpy(datadir, optarg);
         break;
       case '?':               //unknown option --display usage and exit with error
         print_usage ();
         break;
       case ':':               // missing argument--display usage and exit with error
         print_usage ();

       case -1:
         break;                 //all done parsing options
       default:                //not expected to occur
         abort (  );
    };
  } while ( next_option != -1 );
  
  sprintf(chassisfile,"%s/"CHASSISDAT,datadir);
  sprintf(clusterfile,"%s/"CLUSTERDAT,datadir);
  sprintf(nodewatchdat,"%s/"NODEWATCHDAT,datadir);
  sprintf(autonodeconfdat,"%s/"AUTONODECONFDAT,datadir);
  sprintf(nodecommandsdat,"%s/"NODECOMMANDSDAT,datadir);

  mycluster.init(chassisfile,clusterfile,nodecommandsdat);


  if (!nodaemon) {
    switch (fork()) { //detach process
      case -1:
        perror ("fork");
        exit (1);
      case 0:
        break;
      default:
        exit (0);
    }
    setsid ();
  }

  while(true) {
    mycluster.process();
    for(count=1 ; count<10 ; count++) {
      os_hnd->perform_one_op(os_hnd, NULL);
      usleep(5000);
    }
  }
  return 0;
}
 
