#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <netdb.h>
#include <ctype.h>
#include <time.h>

#include <OpenIPMI/ipmiif.h>
#include <OpenIPMI/ipmi_smi.h>
#include <OpenIPMI/ipmi_err.h>
#include <OpenIPMI/ipmi_auth.h>
#include <OpenIPMI/ipmi_lan.h>
#include <OpenIPMI/ipmi_posix.h>

static sdata_t *alloc_sdata(ipmi_sensor_t *sensor)
{
    sdata_t *sdata;

    sdata = (sdata_t *)malloc(sizeof(*sdata));
    if (!sdata)
	return NULL;

    sdata->es = (ipmi_event_state_t *)malloc(ipmi_event_state_size());
    if (!sdata->es) {
	free(sdata);
	return NULL;
    }
    ipmi_event_state_init(sdata->es);

    sdata->th = (ipmi_thresholds_t *)malloc(ipmi_thresholds_size());
    if (!sdata->th) {
	free(sdata->es);
	free(sdata);
	return NULL;
    }
    ipmi_thresholds_init(sdata->th);

    sdata->refcount = 1;

    sdata->sensor_id = ipmi_sensor_convert_to_id(sensor);
    ipmi_sensor_get_name(sensor, sdata->name, sizeof(sdata->name));

    sdata->next = sdata_list;
    sdata->prev = NULL;
    sdata_list = sdata;

    return sdata;
}

/*static sdata_t *find_sdata(ipmi_sensor_t *sensor)
{
    ipmi_sensor_id_t id = ipmi_sensor_convert_to_id(sensor);
    sdata_t          *link;

    link = sdata_list;
    while (link) {
	if (ipmi_cmp_sensor_id(id, link->sensor_id) == 0)
	    return link;
	link = link->next;
    }
    return NULL;
}
*/

static void use_sdata(sdata_t *sdata)
{
    sdata->refcount++;
}

static void release_sdata(sdata_t *sdata)
{
    sdata->refcount--;
    if (sdata->refcount == 0) {
	/* Remove it from the list. */
	if (sdata->next)
	    sdata->next->prev = sdata->prev;

	if (sdata->prev)
	    sdata->prev->next = sdata->next;
	else
	    sdata_list = sdata->next;

	free(sdata->es);
	free(sdata->th);
	free(sdata);
    }
}

static void
got_thresh_reading(ipmi_sensor_t             *sensor,
		   int                       err,
		   enum ipmi_value_present_e value_present,
		   unsigned int              raw_value,
		   double                    val,
		   ipmi_states_t             *states,
		   void                      *cb_data)
{
    sdata_t            *sdata = (sdata_t *)cb_data;
    char nodename[MAXNODENAME],sensorname[MAXSENSORNAME];
    char readingerror[512];
    if (err) {
        ipmi_get_error_string(err,readingerror,511);
	if (nodaemon) fprintf(stderr, "Error 0x%x getting threshold reading for sensor %s:\n%s\n",
	       err, sdata->name, readingerror);
	goto out;
    }

    domainisup=true;     //this will set the global var to true as soon as at least one node is fully up and reporting sensors
                         //setting it again later as more reports come in should be harmless
    
    //fprintf(stderr, "Got threshold reading for sensor %s\n", sdata->name);

    sensor_nodenameonly(nodename, sdata->name);
    sensor_nameonly(sensorname,sdata->name);

    switch (value_present)
    {
    case IPMI_NO_VALUES_PRESENT:
	if (nodaemon) fprintf(stderr,"  no value present\n");
	break;
    case IPMI_BOTH_VALUES_PRESENT:
	{
	    /*const char *percent = "";
	    const char *base;
	    const char *mod_use = "";
	    const char *modifier = "";
	    const char *rate;*/

	/*    base = ipmi_sensor_get_base_unit_string(sensor);
	    if (ipmi_sensor_get_percentage(sensor))
		percent = "%";
	    switch (ipmi_sensor_get_modifier_unit_use(sensor)) {
	    case IPMI_MODIFIER_UNIT_NONE:
		break;
	    case IPMI_MODIFIER_UNIT_BASE_DIV_MOD:
		mod_use = "/";
		modifier = ipmi_sensor_get_modifier_unit_string(sensor);
		break;
	    case IPMI_MODIFIER_UNIT_BASE_MULT_MOD:
		mod_use = "*";
		modifier = ipmi_sensor_get_modifier_unit_string(sensor);
		break;
	    }
	    rate = ipmi_sensor_get_rate_unit_string(sensor);
	  */  
            mycluster.reportsensor(nodename,sensorname,val);
	}
        break;
    case IPMI_RAW_VALUE_PRESENT:
	if (nodaemon) fprintf(stderr,"raw value only (0x%2.2x) not supported\n", raw_value);
    }
  out:
    release_sdata(sdata);
}

/*static void
got_discrete_states(ipmi_sensor_t *sensor,
		    int           err,
		    ipmi_states_t *states,
		    void          *cb_data)
{
    sdata_t *sdata = (sdata_t *)cb_data;
    int     i;

    if (err) {
	printf("Error 0x%x getting discrete states for sensor %s\n",
	       err, sdata->name);
	goto out;
    }

    if (err) {
	printf("Error 0x%x getting discrete states for sensor %s\n",
	       err, sdata->name);
	goto out;
    }

    printf("Got state reading for sensor %s\n", sdata->name);
    if (ipmi_is_event_messages_enabled(states))
	printf("  event messages enabled\n");
    if (ipmi_is_sensor_scanning_enabled(states))
	printf("  sensor scanning enabled\n");
    if (ipmi_is_initial_update_in_progress(states))
	printf("  initial update in progress\n");

    for (i=0; i<15; i++) {
	int val, rv;

	rv = ipmi_sensor_discrete_event_readable(sensor, i, &val);
	if (rv || !val)
	    continue;

	printf("  state %d value is %d\n", i, ipmi_is_state_set(states, i));
    }

 out:
    release_sdata(sdata);
}
*/

bool sensor_nodenameonly(char *nodename, char *fullname) {
  int count=0;
  while(fullname[count]!='\0' && fullname[count]!='(') { 
    nodename[count]=fullname[count];
    count++;
  }
  if(fullname[count]=='\0') return false;
  else {
    nodename[count]='\0';
    return true;
  }
}

bool sensor_nameonly(char *sensorname, char *fullname) {

  int count=0;
  while(fullname[count]!='\0' && fullname[count]!=')') count++;
  if(fullname[count]=='\0') return false;
  count+=2;
  if(fullname[count]=='\0') return false;
  else {
    strcpy(sensorname, &fullname[count]);
    return true;
  }
}

/* Whenever the status of a sensor changes, the function is called
   We display the information of the sensor if we find a new sensor */
static void
sensor_change(enum ipmi_update_e op,
	      ipmi_entity_t      *ent,
	      ipmi_sensor_t      *sensor,
	      void               *cb_data)
{
    sdata_t *sdata;
    char nodename[MAXNODENAME];

    if (op == IPMI_ADDED) {
	sdata = alloc_sdata(sensor);
	if (!sdata) {
	    if (nodaemon) printf("Unable to allocate sensor name memory\n");
	    return;
	}
        sensor_nodenameonly(nodename, sdata->name);

        mycluster.registersensor(nodename,sensor);  //reports the sensor via the cluster
                                                    //class, passing it onto node class
                                                    //if it is a monitored sensor it gets
                                                    //added to the watch list
        release_sdata(sdata);
    }
    return;
}

/* Whenever the status of an entity changes, the function is called
   When a new entity is created, we search all sensors that belong 
   to the entity */
static void
entity_change(enum ipmi_update_e op,
	      ipmi_domain_t      *domain,
	      ipmi_entity_t      *entity,
	      void               *cb_data)
{
    int  rv;
    char name[50];

    ipmi_entity_get_name(entity, name, sizeof(name));
    if (op == IPMI_ADDED) {
	    /* Register callback so that when the status of a
	       sensor changes, sensor_change is called */
	    rv = ipmi_entity_add_sensor_update_handler(entity,
						       sensor_change,
						       NULL);
	    if (rv) {
		printf("ipmi_entity_set_sensor_update_handler: 0x%x", rv);
		exit(1);
	    }
    }
}

/* After we have established connection to domain, this function get called
   At this time, we can do whatever things we want to do. Herr we want to
   search all entities in the system */ 
void
domain_open(ipmi_domain_t *domain,
	   int           err,
	   unsigned int  conn_num,
	   unsigned int  port_num,
	   int           still_connected,
	   void          *user_data)
{
    int rv;
    char domainname[MAXNODENAME];
    char domainerror[512];
    
    if(err!=0) {
      ipmi_domain_get_name(domain,domainname,MAXNODENAME-1);
      mycluster.zerosensors(domainname);
    }

    if (err!=0 && nodaemon) {
      ipmi_get_error_string(err,domainerror,511);
      ipmi_domain_get_name(domain,domainname,MAXNODENAME-1);
      fprintf(stderr,"In domain_open for %s with error:\n%s\n",domainname, domainerror);
    }
   
    /* Register a callback function entity_change. When a new entities 
       is created, entity_change is called */
    rv = ipmi_domain_add_entity_update_handler(domain, entity_change, domain);
    if (rv && nodaemon) {      
	printf("ipmi_domain_add_entity_update_handler return error: %d\n", rv);
	return;
    }
    ipmi_domain_get_name(domain,domainname,MAXNODENAME-1);
    if(domainisup) mycluster.reportdomain(domainname,err);   //notify cluster of current domain status
}

static void
domain_up(ipmi_domain_t *domain, void *cb_data)
{
    return;
    //domainisup=true;     //this will set the global var to true as soon as at least one node is fully up
                         //setting it again later as other nodes come up won't hurt either
}

static void
my_vlog(os_handler_t         *handler,
        const char           *format,
        enum ipmi_log_type_e log_type,
        va_list              ap)
{
  int do_nl = 1;
  if (debugipmilog!=true) return;

  switch(log_type) {
    case IPMI_LOG_INFO:
      printf("INFO: ");
      break;

    case IPMI_LOG_WARNING:
      printf("WARN: ");
      break;

    case IPMI_LOG_SEVERE:
      printf("SEVR: ");
      break;

    case IPMI_LOG_FATAL:
      printf("FATL: ");
      break;

    case IPMI_LOG_ERR_INFO:
      printf("EINF: ");
      break;

    case IPMI_LOG_DEBUG_START:
      do_nl = 0;
      /* FALLTHROUGH */
    case IPMI_LOG_DEBUG:
      printf("DEBG: ");
      break;

    case IPMI_LOG_DEBUG_CONT:
      do_nl = 0;
      /* FALLTHROUGH */
    case IPMI_LOG_DEBUG_END:
      break;
  }

  vprintf(format, ap);

  if (do_nl)
    printf("\n");
}
